################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#                       Closures: Closure models to close the                  #
#                   system of PDEs of the boundary layer equations             #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
import numpy as np

class Closures:
    def __init__(self) -> None:
        pass

    def laminarClosure(theta, H, Ret,Me, name):
        """Laminar closure derived from the Falkner-Skan one-parameter profile family
            Nishida and Drela (1996)
        """
        Hk = (H - 0.29*Me**2)/(1+0.113*Me**2) # Kinematic shape factor
        if name == "airfoil":
            Hk = max(Hk, 1.02)
        else:
            Hk = max(Hk, 1.00005)
        Hk = min(Hk,6)
        Hstar2 = (0.064/(Hk-0.8)+0.251)*Me**2 # Density shape parameter
        delta = min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta)
        Hks = Hk - 4.35
        if Hk <= 4.35:
            dens = Hk + 1
            Hstar = 0.0111*Hks**2/dens - 0.0278*Hks**3/dens + 1.528 -0.0002*(Hks*Hk)**2
        elif Hk > 4.35:
            Hstar = 1.528 + 0.015*Hks**2/Hk
        Hstar = (Hstar + 0.028*Me**2)/(1+0.014*Me**2) # Whitfield's minor additional compressibility correction
        if Hk < 5.5:
            tmp = (5.5-Hk)**3 / (Hk+1)
            Cf = (-0.07 + 0.0727*tmp)/Ret
        elif Hk >= 5.5:
            tmp = 1 - 1/(Hk-4.5)
            Cf = (-0.07 + 0.015*tmp**2) /Ret
        if Hk < 4:
            Cd = ( 0.00205*(4.0-Hk)**5.5 + 0.207 ) * (Hstar/(2*Ret))
        elif Hk >= 4:
            HkCd = (Hk-4)**2
            denCd = 1+0.02*HkCd
            Cd = ( -0.0016*HkCd/denCd + 0.207) * (Hstar/(2*Ret))
        if name == 'wake':
            Cd = 2*(1.10*(1.0-1.0/Hk)**2/Hk)* (Hstar/(2*Ret))
            Cf = 0
        return Cd, Cf, delta, Hstar, Hstar2, Hk


    def turbulentClosure(theta, H, Ct, Ret, Me, name):
        """Turbulent closure derived from Nishida and Drela (1996)
        """
        
        H = max(H, 1.0005)
        # eliminate absurd transients
        Ct = max(min(Ct, 0.30), 0.0000001)
        Hk = (H - 0.29*Me**2)/(1+0.113*Me**2)
        Hk = max(min(Hk, 10), 1.00005)
        Hstar2 = ((0.064/(Hk-0.8))+0.251)*Me**2
        #gam = 1.4 - 1
        #Fc = np.sqrt(1+0.5*gam*Me**2)
        Fc = np.sqrt(1+0.2*Me**2)

        if Ret <= 400:
            H0 = 4
        elif Ret > 400:
            H0 = 3 + (400/Ret)

        Ret = max(Ret, 200)
 
        if Hk <= H0:
            Hstar = ((0.5-4/Ret)*((H0-Hk)/(H0-1))**2)*(1.5/(Hk+0.5))+1.5+4/Ret
        elif Hk > H0:
            Hstar = (Hk-H0)**2*(0.007*np.log(Ret)/(Hk-H0+4/np.log(Ret))**2 + 0.015/Hk) + 1.5 + 4/Ret

        Hstar = (Hstar + 0.028*Me**2)/(1+0.014*Me**2) # Whitfield's minor additional compressibility correction
        #logRt = np.log(Ret/Fc)
        #logRt = max(logRt,3)
        #arg = max(-1.33*Hk, -20)
        Us = (Hstar/2)*(1-4*(Hk-1)/(3*H)) # Equivalent normalized wall slip velocity
        delta = min((3.15+ H + (1.72/(Hk-1)))*theta, 12*theta)
        Ctcon = 0.5/(6.7**2*0.75)

        if name == 'wake':
            Us = min(Us, 0.99995)
            n = 2 # two wake halves
            Cf = 0 # no friction inside the wake
            Hkc = Hk - 1
            Cdw = 0  # wall contribution
            Cdd = (0.995-Us)*Ct**2 # turbulent outer layer contribution
            Cdl = 0.15*(0.995-Us)**2/Ret # laminar stress contribution to outer layer
            Cteq = np.sqrt(4*Hstar*Ctcon*((Hk-1)*Hkc**2)/((1-Us)*(Hk**2)*H)) #Here it is Cteq^0.5
        else:
            if Us > 0.95:
                Us = 0.98
            n = 1
            #Cf = (1/Fc)*(0.3*np.exp(arg)*(logRt/2.3026)**(-1.74-0.31*Hk)+0.00011*(np.tanh(4-(Hk/0.875))-1))
            Cf = 1/Fc*(0.3*np.exp(-1.33*Hk)*(np.log10(Ret/Fc))**(-1.74-0.31*Hk) + 0.00011*(np.tanh(4-Hk/0.875) - 1))
            Hkc = max(Hk-1-18/Ret, 0.01)
            # Dissipation coefficient
            Hmin = 1 + 2.1/np.log(Ret)
            Fl = (Hk-1)/(Hmin-1)
            Dfac = 0.5+0.5*np.tanh(Fl)
            Cdw = 0.5*(Cf*Us) * Dfac
            Cdd = (0.995-Us)*Ct**2
            Cdl = 0.15*(0.995-Us)**2/Ret
            Cteq = np.sqrt(Hstar*Ctcon*((Hk-1)*Hkc**2)/((1-Us)*(Hk**2)*H)) #Here it is Cteq^0.5
        Cd = n*(Cdw+ Cdd + Cdl)
        #Ueq = 1.25/Hk*(Cf/2-((Hk-1)/(6.432*Hk))**2)
        return Cd, Cf, delta, Hstar, Hstar2, Hk, Cteq, Us

    def amplRoutine(Hk, Ret, theta):
        '''Compute the amplitude of the TS waves envelop for the transition
        '''
        Dgr = 0.08
        Hk1 = 3.5
        Hk2 = 4
        Hmi = (1/(Hk-1))
        logRet_crit = 2.492*(1/(Hk-1))**0.43 + 0.7*(np.tanh(14*Hmi-9.24)+1.0) # value at which the amplification starts to grow
        if Ret <=0:
            Ret = 1
        if np.log10(Ret) < logRet_crit - Dgr  :
            Ax = 0
        else:
            Rnorm = (np.log10(Ret) - (logRet_crit-Dgr))/(2*Dgr)
            if Rnorm <=0:
                Rfac = 0
            elif Rnorm >= 1:
                Rfac = 1
            else:
                Rfac = 3*Rnorm**2 - 2*Rnorm**3
            # normal envelope amplification rate
            m = -0.05+2.7*Hmi-5.5*Hmi**2+3*Hmi**3+0.1*np.exp(-20*Hmi)
            arg = 3.87*Hmi-2.52
            dndRet = 0.028*(Hk-1)-0.0345*np.exp(-(arg**2))
            Ax = (m*dndRet/theta)*Rfac
            # Set correction for separated profiles
            if Hk > Hk1:
                Hnorm = (Hk - Hk1)/(Hk2-Hk1)
                if Hnorm >=1:
                    Hfac = 1
                else:
                    Hfac = 3*Hnorm**2 - 2*Hnorm**3
                Ax1 = Ax
                Gro = 0.3+0.35*np.exp(-0.15*(Hk-5))
                Tnr = np.tanh(1.2*(np.log10(Ret)-Gro))
                Ax2 = (0.086*Tnr - 0.25/(Hk-1)**1.5)/theta
                if Ax2 < 0:
                    Ax2 = 0
                Ax = Hfac*Ax2 + (1-Hfac)*Ax1
                Ax = max(Ax, 0)
        return Ax